/*
* @Author: yz.dingjian
* @Description: 
* @Date: 2021/3/9 11:19 AM
* @Last Modified by: yz.dingjian
* @Last Modified time: 2021/3/9 11:19 AM
*/

import 'package:analyzer/dart/ast/visitor.dart';
import 'package:analyzer/dart/ast/ast.dart';

class YZWidgetAstVisitor extends SimpleAstVisitor<Map> {
  // 节点解析通用方法
  Map _safelyVisitNode(AstNode node) {
    if (node != null) {
      return node.accept(this);
    }
    return {};
  }

  /// 遍历节点列表
  List<Map> _safelyVisitNodeList(NodeList<AstNode> nodes,
      {bool isFilter = false}) {
    List<Map> maps = [];
    if (nodes != null) {
      int size = nodes.length;
      for (int i = 0; i < size; i++) {
        var node = nodes[i];
        if (node != null) {
          var res = node.accept(this);
          if (res != null) {
            maps.add(res);
          }
        }
      }
    }
    return maps;
  }

  /// 需要解析的节点类型
  /// visitCompilationUnit
  /// visitFunctionDeclaration
  /// visitFunctionExpression
  /// visitFormalParameterList
  /// visitBlockFunctionBody
  /// visitBlock
  /// visitReturnStatement
  /// visitMethodInvocation
  /// visitSimpleIdentifier
  /// visitArgumentList
  /// visitNamedExpression
  /// visitLabel
  /// visitSimpleStringLiteral
  ///

// 0>解析编译单元
  @override
  Map visitCompilationUnit(CompilationUnit node) {
    print(node.declarations);
    return {'root': _safelyVisitNodeList(node.declarations, isFilter: true)};
  }

// 1> 解析函数定义
  @override
  Map visitFunctionDeclaration(FunctionDeclaration node) {
    String comments = node.beginToken.precedingComments.lexeme;
    String name = node.name.name;
    print(comments);
    print(name);
    comments = comments.replaceAll(' ', '');
    if (comments.startsWith('//@yingzi')) {
      Map resultMap = _safelyVisitNode(node.functionExpression);
      resultMap['name'] = name;
      return resultMap;
    }
    return null;
  }

  @override
  Map visitFunctionExpression(FunctionExpression node) {
    print('visitFunctionExpression===>${node.parameters}');

    /// 解析函数表达式
    /// [node.parameters] 函数参数
    /// [node.body] 函数实现
    return _safelyVisitNode(node.body);
  }

  /// 构造函数参数
  @override
  Map visitFormalParameterList(FormalParameterList node) {
    return {};
  }

  /// 解析代码块
  @override
  Map visitBlockFunctionBody(BlockFunctionBody node) {
    return _safelyVisitNode(node.block);
  }

  /// 构造代码块bloc 结构
  @override
  Map visitBlock(Block node) {
    /// block 中不止一条代码，暂时直解析第一条代码块
    /// 暂时只解析block中的第一条语句，即return的代码

    return {'body': _safelyVisitNode(node.statements.first)};
  }

  /// 方法调用 解析
  /// 这个很关键，各类空间的初始化 基本都是方法的调用
  /// 需要在这里拿到类名，以及参数
  @override
  visitMethodInvocation(MethodInvocation node) {
    Map callee;
    if (node.target != null) {
      node.target.accept(this);
      callee = {
        "type": "MemberExpression",
        "object": _safelyVisitNode(node.target),
        "property": _safelyVisitNode(node.methodName),
      };
    } else {
      callee = _safelyVisitNode(node.methodName);
    }

    Map argumentList = _safelyVisitNode(node.argumentList);

    String functionName = node.methodName.name;
    // 构造方法名，如Image.asset中asset
    String typeString;

    // 被调用者信息
    /*
    "callee": {
      "type": "MemberExpression",
      "object": {
        "type": "Identifier",
        "name": "EdgeInsets"
      },
      "property": {
        "type": "Identifier",
        "name": "fromLTRB"
      }
    }
    */
    if (callee != null) {
      if (callee['type'] != null && callee['type'] == 'MemberExpression') {
        if (callee['object'] != null) {
          functionName = callee['object']['name'];
        }
        if (callee['property'] != null) {
          typeString = callee['property']['name'];
          if (typeString != null) {
            argumentList['yzType'] = typeString;
          }
        }
      }
      if (callee['name'] != null) {
        functionName = callee['name'];
      }
    }

    //  判断是影子封装组件
    //   YZWidgets.yzTagView() => yzTagView;
    if (callee['name'] != null) {
      if (callee['name'] == 'YZWidgets' || callee['name'] == 'YZForm') {
        functionName = typeString;
      }
    }

    return {
      "widgetName": functionName,
      "props": argumentList,
      "callee": callee
    };
  }

  /// 构造标识符定义
  @override
  Map visitSimpleIdentifier(SimpleIdentifier node) {
    return {"type": "Identifier", "name": node.name};
  }

  /// 构造函数调用参数
  @override
  visitArgumentList(ArgumentList node) {
    for (var nodeItem in node.arguments) {
      _safelyVisitNode(nodeItem);
    }

    return {
      "type": "propsList",
      "propsList": _safelyVisitNodeList(node.arguments)
    };
  }

  /// 构造函数表达式，包括函数名，表达式
  @override
  visitNamedExpression(NamedExpression node) {
    ///
    Map name = _safelyVisitNode(node.name);
    Map expression = _safelyVisitNode(node.expression);
    return {'name': name, 'expression': expression};
  }

  @override
  visitLabel(Label node) {
    return _safelyVisitNode(node.label);
  }

  @override
  Map visitReturnStatement(ReturnStatement node) {
    return _safelyVisitNode(node.expression);
  }

//  EdgeInsets.fromLTRB(2, 3, 4, 5),
  @override
  visitPrefixedIdentifier(PrefixedIdentifier node) {
    return {
      "type": "PrefixedIdentifier",
      "identifier": _safelyVisitNode(node.identifier),
      "prefix": _safelyVisitNode(node.prefix),
    };
  }

  /// 参数列表，参数是一个数组 例如Children
  @override
  Map visitListLiteral(ListLiteral node) {
    return {'list': _safelyVisitNodeList(node.elements)};
  }

  @override
  Map visitFunctionDeclarationStatement(FunctionDeclarationStatement node) {
    return _safelyVisitNode(node.functionDeclaration);
  }

  @override
  visitSimpleStringLiteral(SimpleStringLiteral node) {
    return {"type": "StringLiteral", "value": node.value};
  }

  @override
  visitBooleanLiteral(BooleanLiteral node) {
    return {"type": "BooleanLiteral", "value": node.value};
  }

  @override
  Map visitIntegerLiteral(IntegerLiteral node) {
    return {"type": "IntegerLiteral", "value": node.value};
  }
}
